import atexit
import io
import json
import logging
import schwab
import unittest

from schwab.client import Client
from .utils import MockResponse, no_duplicates
from unittest.mock import Mock, patch


class RedactorTest(unittest.TestCase):

    def setUp(self):
        self.redactor = schwab.debug.LogRedactor()

    @no_duplicates
    def test_no_redactions(self):
        self.assertEqual('test message', self.redactor.redact('test message'))

    @no_duplicates
    def test_simple_redaction(self):
        self.redactor.register('secret', 'SECRET')

        self.assertEqual(
            '<REDACTED SECRET> message',
            self.redactor.redact('secret message'))

    @no_duplicates
    def test_multiple_registrations_same_string(self):
        self.redactor.register('secret', 'SECRET')
        self.redactor.register('secret', 'SECRET')

        self.assertEqual(
            '<REDACTED SECRET> message',
            self.redactor.redact('secret message'))

    @no_duplicates
    def test_multiple_registrations_same_string_different_label(self):
        self.redactor.register('secret-A', 'SECRET')
        self.redactor.register('secret-B', 'SECRET')

        self.assertEqual(
            '<REDACTED SECRET-1> message <REDACTED SECRET-2>',
            self.redactor.redact('secret-A message secret-B'))


class RegisterRedactionsTest(unittest.TestCase):

    def setUp(self):
        self.captured = io.StringIO()
        self.logger = logging.getLogger('test')
        self.dump_logs = schwab.debug._enable_bug_report_logging(
            output=self.captured, loggers=[self.logger])
        schwab.LOG_REDACTOR = schwab.debug.LogRedactor()

    @no_duplicates
    def test_empty_string(self):
        schwab.debug.register_redactions('')

    @no_duplicates
    def test_empty_dict(self):
        schwab.debug.register_redactions({})

    @no_duplicates
    def test_empty_list(self):
        schwab.debug.register_redactions([])

    @no_duplicates
    def test_dict(self):
        schwab.debug.register_redactions(
            {'BadNumber': '100001'},
            bad_patterns=['bad'])
        schwab.debug.register_redactions(
            {'OtherBadNumber': '200002'},
            bad_patterns=['bad'])

        self.logger.info('Bad Number: 100001')
        self.logger.info('Other Bad Number: 200002')

        self.dump_logs()
        self.assertRegex(
            self.captured.getvalue(),
            r'\[.*\] Bad Number: <REDACTED BadNumber>\n' +
            r'\[.*\] Other Bad Number: <REDACTED OtherBadNumber>\n')

    @no_duplicates
    def test_list_of_dict(self):
        schwab.debug.register_redactions(
            [{'GoodNumber': '900009'},
             {'BadNumber': '100001'},
             {'OtherBadNumber': '200002'}],
            bad_patterns=['bad'])

        self.logger.info('Good Number: 900009')
        self.logger.info('Bad Number: 100001')
        self.logger.info('Other Bad Number: 200002')

        self.dump_logs()
        self.assertRegex(
            self.captured.getvalue(),
            r'\[.*\] Good Number: 900009\n' +
            r'\[.*\] Bad Number: <REDACTED 1-BadNumber>\n' +
            r'\[.*\] Other Bad Number: <REDACTED 2-OtherBadNumber>\n')

    @no_duplicates
    def test_whitelist(self):
        schwab.debug.register_redactions(
            [{'GoodNumber': '900009'},
             {'BadNumber': '100001'},
             {'OtherBadNumber': '200002'}],
            bad_patterns=['bad'],
            whitelisted=['otherbadnumber'])

        self.logger.info('Good Number: 900009')
        self.logger.info('Bad Number: 100001')
        self.logger.info('Other Bad Number: 200002')

        self.dump_logs()
        self.assertRegex(
            self.captured.getvalue(),
            r'\[.*\] Good Number: 900009\n' +
            r'\[.*\] Bad Number: <REDACTED 1-BadNumber>\n' +
            r'\[.*\] Other Bad Number: 200002\n')

    @no_duplicates
    @patch('schwab.debug.register_redactions', new_callable=Mock)
    def test_register_from_request_success(self, register_redactions):
        resp = MockResponse({'success': 1}, 200)
        schwab.debug.register_redactions_from_response(resp)
        register_redactions.assert_called_with({'success': 1})

    @no_duplicates
    @patch('schwab.debug.register_redactions', new_callable=Mock)
    def test_register_from_request_not_okay(self, register_redactions):
        resp = MockResponse({'success': 1}, 403)
        schwab.debug.register_redactions_from_response(resp)
        register_redactions.assert_not_called()

    @no_duplicates
    @patch('schwab.debug.register_redactions', new_callable=Mock)
    def test_register_unparseable_json(self, register_redactions):
        class MR(MockResponse):
            def json(self):
                raise json.decoder.JSONDecodeError('e243rschwabgew', '', 0)

        resp = MR({'success': 1}, 200)
        schwab.debug.register_redactions_from_response(resp)
        register_redactions.assert_not_called()

class EnableDebugLoggingTest(unittest.TestCase):

    @patch('logging.Logger.addHandler')
    def test_enable_doesnt_throw_exceptions(self, _):
        try:
            schwab.debug.enable_bug_report_logging()
        except AttributeError:
            self.fail("debug.enable_bug_report_logging() raised AttributeError unexpectedly")



import json
from schwab.streaming import StreamJsonDecoder


class HeuristicJsonDecoder(StreamJsonDecoder):
    def decode_json_string(self, raw):
        '''
        Attempts the following, in order:
        
        1. Return the JSON decoding of the raw string. 
        2. Replace all instances of ``\\\\\\\\`` with ``\\\\`` and return the 
           decoding.

        Note alternative (and potentially expensive) transformations are only 
        performed when ``JSONDecodeError`` exceptions are raised by earlier 
        stages.
        '''

        # Note "no cover" pragmas are added pending addition of real-world test 
        # cases which trigger this issue.

        try:
            return json.loads(raw)
        except json.decoder.JSONDecodeError:  # pragma: no cover
            raw = raw.replace('\\\\', '\\')

        return json.loads(raw)  # pragma: no cover



import httpx
import unittest

from schwab.orders.generic import *
from schwab.orders.common import *
from ..utils import has_diff, no_duplicates


class OrderBuilderTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.order_builder = OrderBuilder()

    ##########################################################################
    # Session

    @no_duplicates
    def test_session_success(self):
        self.order_builder.set_session(Session.NORMAL)
        self.assertFalse(has_diff({
            'session': 'NORMAL'
        }, self.order_builder.build()))

        self.order_builder.clear_session()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_session_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.Session.NORMAL'):
            self.order_builder.set_session('NORMAL')

    @no_duplicates
    def test_session_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_session('NORMAL')
        self.assertFalse(has_diff({
            'session': 'NORMAL'
        }, self.order_builder.build()))

    ##########################################################################
    # Duration

    @no_duplicates
    def test_duration_success(self):
        self.order_builder.set_duration(Duration.DAY)
        self.assertFalse(has_diff({
            'duration': 'DAY'
        }, self.order_builder.build()))

        self.order_builder.clear_duration()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_duration_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.Duration.DAY'):
            self.order_builder.set_duration('DAY')

    @no_duplicates
    def test_duration_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_duration('DAY')
        self.assertFalse(has_diff({
            'duration': 'DAY'
        }, self.order_builder.build()))

    ##########################################################################
    # OrderType

    @no_duplicates
    def test_order_type_success(self):
        self.order_builder.set_order_type(OrderType.MARKET)
        self.assertFalse(has_diff({
            'orderType': 'MARKET'
        }, self.order_builder.build()))

        self.order_builder.clear_order_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_order_type_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.OrderType.MARKET'):
            self.order_builder.set_order_type('MARKET')

    @no_duplicates
    def test_order_type_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_order_type('MARKET')
        self.assertFalse(has_diff({
            'orderType': 'MARKET'
        }, self.order_builder.build()))

    ##########################################################################
    # ComplexOrderStrategyType

    @no_duplicates
    def test_complex_order_strategy_type_success(self):
        self.order_builder.set_complex_order_strategy_type(
            ComplexOrderStrategyType.IRON_CONDOR)
        self.assertFalse(has_diff({
            'complexOrderStrategyType': 'IRON_CONDOR'
        }, self.order_builder.build()))

        self.order_builder.clear_complex_order_strategy_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test__wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 
                'schwab.orders.common.ComplexOrderStrategyType.IRON_CONDOR'):
            self.order_builder.set_complex_order_strategy_type('IRON_CONDOR')

    @no_duplicates
    def test__wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_complex_order_strategy_type('IRON_CONDOR')
        self.assertFalse(has_diff({
            'complexOrderStrategyType': 'IRON_CONDOR'
        }, self.order_builder.build()))

    ##########################################################################
    # Quantity

    @no_duplicates
    def test_quantity_success(self):
        self.order_builder.set_quantity(12)
        self.assertFalse(has_diff({
            'quantity': 12
        }, self.order_builder.build()))

        self.order_builder.clear_quantity()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_quantity_negative(self):
        with self.assertRaises(ValueError):
            self.order_builder.set_quantity(-12)

    @no_duplicates
    def test_quantity_zero(self):
        with self.assertRaises(ValueError):
            self.order_builder.set_quantity(0)

    ##########################################################################
    # RequestedDestination

    @no_duplicates
    def test_destination_link_name_success(self):
        self.order_builder.set_destination_link_name(Destination.INET)
        self.assertFalse(has_diff({
            'destinationLinkName': 'INET'
        }, self.order_builder.build()))

        self.order_builder.clear_destination_link_name()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_destination_link_name_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.Destination.INET'):
            self.order_builder.set_destination_link_name('INET')

    @no_duplicates
    def test_destination_link_name_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_destination_link_name('INET')
        self.assertFalse(has_diff({
            'destinationLinkName': 'INET'
        }, self.order_builder.build()))

    ##########################################################################
    # StopPrice

    @no_duplicates
    def test_stop_price_success(self):
        self.order_builder.set_stop_price(42.90)
        self.assertFalse(has_diff({
            'stopPrice': '42.90'
        }, self.order_builder.build()))

        self.order_builder.clear_stop_price()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_as_string(self):
        self.order_builder.set_stop_price('invalid')
        self.assertFalse(has_diff({
            'stopPrice': 'invalid'
        }, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_negative(self):
        self.order_builder.set_stop_price(-1.31)
        self.assertFalse(has_diff({
            'stopPrice': '-1.31'
        }, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_zero(self):
        self.order_builder.set_stop_price(0)
        self.assertFalse(has_diff({
            'stopPrice': '0.00'
        }, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_do_not_round_up(self):
        self.order_builder.set_stop_price(1.99999)
        self.assertFalse(has_diff({
            'stopPrice': '1.99'
        }, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_do_round_down(self):
        self.order_builder.set_stop_price(2.00001)
        self.assertFalse(has_diff({
            'stopPrice': '2.00'
        }, self.order_builder.build()))

    @no_duplicates
    def test_copy_stop_price(self):
        self.order_builder.copy_stop_price(2.00001)
        self.assertFalse(has_diff({
            'stopPrice': 2.00001
        }, self.order_builder.build()))

    @no_duplicates
    def test_copy_stop_price_hopelessly_invalid(self):
        self.order_builder.copy_stop_price(['hopelessly invalid'])
        self.assertFalse(has_diff({
            'stopPrice': ['hopelessly invalid']
        }, self.order_builder.build()))

    ##########################################################################
    # StopPriceLinkBasis

    @no_duplicates
    def test_stop_price_link_basis_success(self):
        self.order_builder.set_stop_price_link_basis(StopPriceLinkBasis.ASK)
        self.assertFalse(has_diff({
            'stopPriceLinkBasis': 'ASK'
        }, self.order_builder.build()))

        self.order_builder.clear_stop_price_link_basis()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_link_basis_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.StopPriceLinkBasis.ASK'):
            self.order_builder.set_stop_price_link_basis('ASK')

    @no_duplicates
    def test_stop_price_link_basis_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_stop_price_link_basis('ASK')
        self.assertFalse(has_diff({
            'stopPriceLinkBasis': 'ASK'
        }, self.order_builder.build()))

    ##########################################################################
    # StopPriceLinkType

    @no_duplicates
    def test_stop_price_link_type_success(self):
        self.order_builder.set_stop_price_link_type(StopPriceLinkType.VALUE)
        self.assertFalse(has_diff({
            'stopPriceLinkType': 'VALUE'
        }, self.order_builder.build()))

        self.order_builder.clear_stop_price_link_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_stop_price_link_type_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.StopPriceLinkType.VALUE'):
            self.order_builder.set_stop_price_link_type('VALUE')

    @no_duplicates
    def test_stop_price_link_type_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_stop_price_link_type('VALUE')
        self.assertFalse(has_diff({
            'stopPriceLinkType': 'VALUE'
        }, self.order_builder.build()))

    ##########################################################################
    # StopPriceOffset

    @no_duplicates
    def test_stop_price_offset_success(self):
        self.order_builder.set_stop_price_offset(12.98)
        self.assertFalse(has_diff({
            'stopPriceOffset': 12.98
        }, self.order_builder.build()))

        self.order_builder.clear_stop_price_offset()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    ##########################################################################
    # StopType

    @no_duplicates
    def test_stop_type_success(self):
        self.order_builder.set_stop_type(StopType.MARK)
        self.assertFalse(has_diff({
            'stopType': 'MARK'
        }, self.order_builder.build()))

        self.order_builder.clear_stop_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_stop_type_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.StopType.MARK'):
            self.order_builder.set_stop_type('MARK')

    @no_duplicates
    def test_stop_type_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_stop_type('MARK')
        self.assertFalse(has_diff({
            'stopType': 'MARK'
        }, self.order_builder.build()))

    ##########################################################################
    # PriceLinkBasis

    @no_duplicates
    def test_price_link_basis_success(self):
        self.order_builder.set_price_link_basis(PriceLinkBasis.AVERAGE)
        self.assertFalse(has_diff({
            'priceLinkBasis': 'AVERAGE'
        }, self.order_builder.build()))

        self.order_builder.clear_price_link_basis()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_price_link_basis_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.PriceLinkBasis.AVERAGE'):
            self.order_builder.set_price_link_basis('AVERAGE')

    @no_duplicates
    def test_price_link_basis_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_price_link_basis('AVERAGE')
        self.assertFalse(has_diff({
            'priceLinkBasis': 'AVERAGE'
        }, self.order_builder.build()))

    ##########################################################################
    # PriceLinkType

    @no_duplicates
    def test_price_link_type_success(self):
        self.order_builder.set_price_link_type(PriceLinkType.PERCENT)
        self.assertFalse(has_diff({
            'priceLinkType': 'PERCENT'
        }, self.order_builder.build()))

        self.order_builder.clear_price_link_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_price_link_type_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.PriceLinkType.PERCENT'):
            self.order_builder.set_price_link_type('PERCENT')

    @no_duplicates
    def test_price_link_type_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_price_link_type('PERCENT')
        self.assertFalse(has_diff({
            'priceLinkType': 'PERCENT'
        }, self.order_builder.build()))

    ##########################################################################
    # Price

    @no_duplicates
    def test_price_success(self):
        self.order_builder.set_price(23.49)
        self.assertFalse(has_diff({
            'price': '23.49'
        }, self.order_builder.build()))

        self.order_builder.clear_price()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_price_success_as_string(self):
        self.order_builder.set_price('invalid')
        self.assertFalse(has_diff({
            'price': 'invalid'
        }, self.order_builder.build()))

    @no_duplicates
    def test_price_negative(self):
        self.order_builder.set_price(-1.23)
        self.assertFalse(has_diff({
            'price': '-1.23'
        }, self.order_builder.build()))

    @no_duplicates
    def test_price_zero(self):
        self.order_builder.set_price(0.0)
        self.assertFalse(has_diff({
            'price': '0.00'
        }, self.order_builder.build()))

    @no_duplicates
    def test_price_do_not_round_up(self):
        self.order_builder.set_price(19.9999999)
        self.assertFalse(has_diff({
            'price': '19.99'
        }, self.order_builder.build()))

    @no_duplicates
    def test_price_do_not_round_down(self):
        self.order_builder.set_price(20.00000001)
        self.assertFalse(has_diff({
            'price': '20.00'
        }, self.order_builder.build()))

    @no_duplicates
    def test_copy_price(self):
        self.order_builder.copy_price(19.9999999)
        self.assertFalse(has_diff({
            'price': 19.9999999
        }, self.order_builder.build()))

    @no_duplicates
    def test_copy_price_hopelessly_invalid(self):
        self.order_builder.copy_price(['hopelessly invalid'])
        self.assertFalse(has_diff({
            'price': ['hopelessly invalid']
        }, self.order_builder.build()))

    ##########################################################################
    # ActivationPrice

    @no_duplicates
    def test_activation_price_success(self):
        self.order_builder.set_activation_price(54.03)
        self.assertFalse(has_diff({
            'activationPrice': 54.03
        }, self.order_builder.build()))

        self.order_builder.clear_activation_price()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_activation_price_negative(self):
        with self.assertRaises(ValueError):
            self.order_builder.set_activation_price(-3.94)

    @no_duplicates
    def test_activation_price_zero(self):
        with self.assertRaises(ValueError):
            self.order_builder.set_activation_price(0.0)

    ##########################################################################
    # SpecialInstruction

    @no_duplicates
    def test_special_instruction_success(self):
        self.order_builder.set_special_instruction(
            SpecialInstruction.DO_NOT_REDUCE)
        self.assertFalse(has_diff({
            'specialInstruction': 'DO_NOT_REDUCE'
        }, self.order_builder.build()))

        self.order_builder.clear_special_instruction()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_special_instruction_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError,
                'schwab.orders.common.SpecialInstruction.DO_NOT_REDUCE'):
            self.order_builder.set_special_instruction('DO_NOT_REDUCE')

    @no_duplicates
    def test_special_instruction_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_special_instruction('DO_NOT_REDUCE')
        self.assertFalse(has_diff({
            'specialInstruction': 'DO_NOT_REDUCE'
        }, self.order_builder.build()))

    ##########################################################################
    # OrderStrategyType

    @no_duplicates
    def test_order_strategy_type_success(self):
        self.order_builder.set_order_strategy_type(OrderStrategyType.OCO)
        self.assertFalse(has_diff({
            'orderStrategyType': 'OCO'
        }, self.order_builder.build()))

        self.order_builder.clear_order_strategy_type()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_order_strategy_type_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.OrderStrategyType.OCO'):
            self.order_builder.set_order_strategy_type('OCO')

    @no_duplicates
    def test_order_strategy_type_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_order_strategy_type('OCO')
        self.assertFalse(has_diff({
            'orderStrategyType': 'OCO'
        }, self.order_builder.build()))

    ##########################################################################
    # ChildOrderStrategies

    @no_duplicates
    def test_add_child_order_strategy_success(self):
        self.order_builder.add_child_order_strategy(
            OrderBuilder().set_session(Session.NORMAL))
        self.assertFalse(has_diff({
            'childOrderStrategies': [{'session': 'NORMAL'}]
        }, self.order_builder.build()))

        self.order_builder.clear_child_order_strategies()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_add_child_order_strategy_dict(self):
        self.order_builder.add_child_order_strategy(
            {'session': 'NORMAL'})
        self.assertFalse(has_diff({
            'childOrderStrategies': [{'session': 'NORMAL'}]
        }, self.order_builder.build()))

    @no_duplicates
    def test_add_child_order_strategy_invalid_type(self):
        with self.assertRaises(ValueError):
            self.order_builder.add_child_order_strategy(10)

    @no_duplicates
    def test_add_child_order_strategy_httpx_response(self):
        with self.assertRaisesRegex(
                ValueError, 'Child order cannot be a response'):
            self.order_builder.add_child_order_strategy(httpx.Response(200))

    ##########################################################################
    # OrderLegCollection

    @no_duplicates
    def test_add_equity_leg_success(self):
        self.order_builder.add_equity_leg(EquityInstruction.BUY, 'GOOG', 10)
        self.order_builder.add_equity_leg(
            EquityInstruction.SELL_SHORT, 'MSFT', 1)
        self.assertFalse(has_diff({
            'orderLegCollection': [{
                'instruction': 'BUY',
                'instrument': {
                    'symbol': 'GOOG',
                    'assetType': 'EQUITY'
                },
                'quantity': 10,
            }, {
                'instruction': 'SELL_SHORT',
                'instrument': {
                    'symbol': 'MSFT',
                    'assetType': 'EQUITY'
                },
                'quantity': 1,
            }]
        }, self.order_builder.build()))

        self.order_builder.clear_order_legs()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_add_equity_leg_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.EquityInstruction.BUY'):
            self.order_builder.add_equity_leg('BUY', 'GOOG', 10)

    @no_duplicates
    def test_add_equity_leg_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)

        self.order_builder.add_equity_leg('BUY', 'GOOG', 10)
        self.order_builder.add_equity_leg('SELL_TO_CLOSE', 'MSFT', 1)

        self.assertFalse(has_diff({
            'orderLegCollection': [{
                'instruction': 'BUY',
                'instrument': {
                    'symbol': 'GOOG',
                    'assetType': 'EQUITY'
                },
                'quantity': 10,
            }, {
                'instruction': 'SELL_TO_CLOSE',
                'instrument': {
                    'symbol': 'MSFT',
                    'assetType': 'EQUITY'
                },
                'quantity': 1,
            }]
        }, self.order_builder.build()))

    @no_duplicates
    def test_add_equity_leg_negative_quantity(self):
        with self.assertRaises(ValueError):
            self.order_builder.add_equity_leg(
                EquityInstruction.BUY, 'GOOG', -1)

    @no_duplicates
    def test_add_equity_leg_zero_quantity(self):
        with self.assertRaises(ValueError):
            self.order_builder.add_equity_leg(
                EquityInstruction.BUY, 'GOOG', 0)

    @no_duplicates
    def test_add_option_leg_success(self):
        self.order_builder.add_option_leg(
            OptionInstruction.BUY_TO_OPEN, 'GOOG31433C1342', 10)
        self.order_builder.add_option_leg(
            OptionInstruction.BUY_TO_CLOSE, 'MSFT439132P35', 1)
        self.assertFalse(has_diff({
            'orderLegCollection': [{
                'instruction': 'BUY_TO_OPEN',
                'instrument': {
                    'symbol': 'GOOG31433C1342',
                    'assetType': 'OPTION'
                },
                'quantity': 10,
            }, {
                'instruction': 'BUY_TO_CLOSE',
                'instrument': {
                    'symbol': 'MSFT439132P35',
                    'assetType': 'OPTION'
                },
                'quantity': 1,
            }]
        }, self.order_builder.build()))

        self.order_builder.clear_order_legs()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_add_option_leg_wrong_type(self):
        with self.assertRaisesRegex(
                ValueError, 'schwab.orders.common.OptionInstruction.BUY_TO_OPEN'):
            self.order_builder.add_option_leg(
                'BUY_TO_OPEN', 'GOOG31433C1342', 10)

    @no_duplicates
    def test_add_option_leg_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)

        self.order_builder.add_option_leg('BUY_TO_OPEN', 'GOOG31433C1342', 10)
        self.order_builder.add_option_leg('BUY_TO_CLOSE', 'MSFT439132P35', 1)
        self.assertFalse(has_diff({
            'orderLegCollection': [{
                'instruction': 'BUY_TO_OPEN',
                'instrument': {
                    'symbol': 'GOOG31433C1342',
                    'assetType': 'OPTION'
                },
                'quantity': 10,
            }, {
                'instruction': 'BUY_TO_CLOSE',
                'instrument': {
                    'symbol': 'MSFT439132P35',
                    'assetType': 'OPTION'
                },
                'quantity': 1,
            }]
        }, self.order_builder.build()))

        self.order_builder.clear_order_legs()
        self.assertFalse(has_diff({}, self.order_builder.build()))

    @no_duplicates
    def test_add_option_leg_negative_quantity(self):
        with self.assertRaises(ValueError):
            self.order_builder.add_option_leg(
                OptionInstruction.BUY_TO_OPEN, 'GOOG31433C1342', -1)

    @no_duplicates
    def test_add_option_leg_zero_quantity(self):
        with self.assertRaises(ValueError):
            self.order_builder.add_option_leg(
                OptionInstruction.BUY_TO_OPEN, 'GOOG31433C1342', 0)


class OrderBuilderExamplesTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None

    ##########################################################################
    # Functional tests from here:
    # https://developer.tdameritrade.com/content/place-order-samples
    @no_duplicates
    def test_quantity_negative(self):
        with self.assertRaises(ValueError):
            self.order_builder.set_quantity(-12)

    @no_duplicates
    def test_quantity_wrong_type_no_check(self):
        self.order_builder = OrderBuilder(enforce_enums=False)
        self.order_builder.set_quantity('')
        self.assertFalse(has_diff({
            'quantity': ''
        }, self.order_builder.build()))


class OrderBuilderExamplesTest(unittest.TestCase):

    @no_duplicates
    def setUp(self):
        self.maxDiff = None

    ##########################################################################
    # Functional tests from here:
    # https://developer.tdameritrade.com/content/place-order-samples

    @no_duplicates
    def test_buy_market_stock(self):
        builder = (
            OrderBuilder()
            .set_order_type(OrderType.MARKET)
            .set_session(Session.NORMAL)
            .set_duration(Duration.DAY)
            .set_order_strategy_type(OrderStrategyType.SINGLE)
            .add_equity_leg(EquityInstruction.BUY, 'XYZ', 15))

        expected = {
            'orderType': 'MARKET',
            'session': 'NORMAL',
            'duration': 'DAY',
            'orderStrategyType': 'SINGLE',
            'orderLegCollection': [
                {
                    'instruction': 'BUY',  # The original says 'Buy'
                    'quantity': 15,
                    'instrument': {
                        'symbol': 'XYZ',
                        'assetType': 'EQUITY'
                    }
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_buy_limit_single_option(self):
        builder = (
            OrderBuilder()
            .set_complex_order_strategy_type(ComplexOrderStrategyType.NONE)
            .set_order_type(OrderType.LIMIT)
            .set_session(Session.NORMAL)
            .set_price(6.45)
            .set_duration(Duration.DAY)
            .set_order_strategy_type(OrderStrategyType.SINGLE)
            .add_option_leg(OptionInstruction.BUY_TO_OPEN, 'XYZ_032015C49', 10))

        expected = {
            'complexOrderStrategyType': 'NONE',
            'orderType': 'LIMIT',
            'session': 'NORMAL',
            'price': '6.45',
            'duration': 'DAY',
            'orderStrategyType': 'SINGLE',
            'orderLegCollection': [
                {
                    'instruction': 'BUY_TO_OPEN',
                    'quantity': 10,
                    'instrument': {
                        'symbol': 'XYZ_032015C49',
                        'assetType': 'OPTION'
                    }
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_buy_limit_vertical_call_spread(self):
        builder = (
            OrderBuilder()
            .set_order_type(OrderType.NET_DEBIT)
            .set_session(Session.NORMAL)
            .set_price(1.20)
            .set_duration(Duration.DAY)
            .set_order_strategy_type(OrderStrategyType.SINGLE)
            .add_option_leg(OptionInstruction.BUY_TO_OPEN, 'XYZ_011516C40', 10)
            .add_option_leg(
                OptionInstruction.SELL_TO_OPEN, 'XYZ_011516C42.5', 10))

        expected = {
            'orderType': 'NET_DEBIT',
            'session': 'NORMAL',
            'price': '1.20',
            'duration': 'DAY',
            'orderStrategyType': 'SINGLE',
            'orderLegCollection': [
                {
                    'instruction': 'BUY_TO_OPEN',
                    'quantity': 10,
                    'instrument': {
                        'symbol': 'XYZ_011516C40',
                        'assetType': 'OPTION'
                    }
                },
                {
                    'instruction': 'SELL_TO_OPEN',
                    'quantity': 10,
                    'instrument': {
                        'symbol': 'XYZ_011516C42.5',
                        'assetType': 'OPTION'
                    }
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_custom_option_spread(self):
        builder = (
            OrderBuilder()
            .set_order_strategy_type(OrderStrategyType.SINGLE)
            .set_order_type(OrderType.MARKET)
            .add_option_leg(OptionInstruction.SELL_TO_OPEN, 'XYZ_011819P45', 1)
            .add_option_leg(OptionInstruction.BUY_TO_OPEN, 'XYZ_011720P43', 2)
            .set_complex_order_strategy_type(ComplexOrderStrategyType.CUSTOM)
            .set_duration(Duration.DAY)
            .set_session(Session.NORMAL))

        expected = {
            'orderStrategyType': 'SINGLE',
            'orderType': 'MARKET',
            'orderLegCollection': [
                {
                    'instrument': {
                        'assetType': 'OPTION',
                        'symbol': 'XYZ_011819P45'
                    },
                    'instruction': 'SELL_TO_OPEN',
                    'quantity': 1
                },
                {
                    'instrument': {
                        'assetType': 'OPTION',
                        'symbol': 'XYZ_011720P43'
                    },
                    'instruction': 'BUY_TO_OPEN',
                    'quantity': 2
                }
            ],
            'complexOrderStrategyType': 'CUSTOM',
            'duration': 'DAY',
            'session': 'NORMAL'
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_conditional_order_one_triggers_another(self):
        builder = (
            OrderBuilder()
            .set_order_type(OrderType.LIMIT)
            .set_session(Session.NORMAL)
            .set_price(34.97)
            .set_duration(Duration.DAY)
            .set_order_strategy_type(OrderStrategyType.TRIGGER)
            .add_equity_leg(EquityInstruction.BUY, 'XYZ', 10)
            .add_child_order_strategy(
                OrderBuilder()
                .set_order_type(OrderType.LIMIT)
                .set_session(Session.NORMAL)
                .set_price(42.03)
                .set_duration(Duration.DAY)
                .set_order_strategy_type(OrderStrategyType.SINGLE)
                .add_equity_leg(EquityInstruction.SELL, 'XYZ', 10)))

        expected = {
            'orderType': 'LIMIT',
            'session': 'NORMAL',
            'price': '34.97',
            'duration': 'DAY',
            'orderStrategyType': 'TRIGGER',
            'orderLegCollection': [
                {
                    'instruction': 'BUY',
                    'quantity': 10,
                    'instrument': {
                        'symbol': 'XYZ',
                        'assetType': 'EQUITY'
                    }
                }
            ],
            'childOrderStrategies': [
                {
                    'orderType': 'LIMIT',
                    'session': 'NORMAL',
                    'price': '42.03',
                    'duration': 'DAY',
                    'orderStrategyType': 'SINGLE',
                    'orderLegCollection': [
                        {
                            'instruction': 'SELL',
                            'quantity': 10,
                            'instrument': {
                                'symbol': 'XYZ',
                                'assetType': 'EQUITY'
                            }
                        }
                    ]
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_conditional_order_one_cancels_another(self):
        builder = (
            OrderBuilder()
            .set_order_strategy_type(OrderStrategyType.OCO)
            .add_child_order_strategy(
                OrderBuilder()
                .set_order_type(OrderType.LIMIT)
                .set_session(Session.NORMAL)
                .set_price(45.97)
                .set_duration(Duration.DAY)
                .set_order_strategy_type(OrderStrategyType.SINGLE)
                .add_equity_leg(EquityInstruction.SELL, 'XYZ', 2)
            )
            .add_child_order_strategy(
                OrderBuilder()
                .set_order_type(OrderType.STOP_LIMIT)
                .set_session(Session.NORMAL)
                .set_price(37.00)
                .set_stop_price(37.03)
                .set_duration(Duration.DAY)
                .set_order_strategy_type(OrderStrategyType.SINGLE)
                .add_equity_leg(EquityInstruction.SELL, 'XYZ', 2)))

        expected = {
            'orderStrategyType': 'OCO',
            'childOrderStrategies': [
                {
                    'orderType': 'LIMIT',
                    'session': 'NORMAL',
                    'price': '45.97',
                    'duration': 'DAY',
                    'orderStrategyType': 'SINGLE',
                    'orderLegCollection': [
                        {
                            'instruction': 'SELL',
                            'quantity': 2,
                            'instrument': {
                                'symbol': 'XYZ',
                                'assetType': 'EQUITY'
                            }
                        }
                    ]
                },
                {
                    'orderType': 'STOP_LIMIT',
                    'session': 'NORMAL',
                    'price': '37.00',
                    'stopPrice': '37.03',
                    'duration': 'DAY',
                    'orderStrategyType': 'SINGLE',
                    'orderLegCollection': [
                        {
                            'instruction': 'SELL',
                            'quantity': 2,
                            'instrument': {
                                'symbol': 'XYZ',
                                'assetType': 'EQUITY'
                            }
                        }
                    ]
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_conditional_order_one_triggers_a_one_cancels_other(self):
        builder = (
            OrderBuilder()
            .set_order_strategy_type(OrderStrategyType.TRIGGER)
            .set_session(Session.NORMAL)
            .set_duration(Duration.DAY)
            .set_order_type(OrderType.LIMIT)
            .set_price(14.97)
            .add_equity_leg(EquityInstruction.BUY, 'XYZ', 5)
            .add_child_order_strategy(
                OrderBuilder()
                .set_order_strategy_type(OrderStrategyType.OCO)
                .add_child_order_strategy(
                    OrderBuilder()
                    .set_order_strategy_type(OrderStrategyType.SINGLE)
                    .set_session(Session.NORMAL)
                    .set_duration(Duration.GOOD_TILL_CANCEL)
                    .set_order_type(OrderType.LIMIT)
                    .set_price(15.27)
                    .add_equity_leg(EquityInstruction.SELL, 'XYZ', 5))
                .add_child_order_strategy(
                    OrderBuilder()
                    .set_order_strategy_type(OrderStrategyType.SINGLE)
                    .set_session(Session.NORMAL)
                    .set_duration(Duration.GOOD_TILL_CANCEL)
                    .set_order_type(OrderType.STOP)
                    .set_stop_price(11.27)
                    .add_equity_leg(EquityInstruction.SELL, 'XYZ', 5))))

        expected = {
            'orderStrategyType': 'TRIGGER',
            'session': 'NORMAL',
            'duration': 'DAY',
            'orderType': 'LIMIT',
            'price': '14.97',
            'orderLegCollection': [
                {
                    'instruction': 'BUY',
                    'quantity': 5,
                    'instrument': {
                        'assetType': 'EQUITY',
                        'symbol': 'XYZ'
                    }
                }
            ],
            'childOrderStrategies': [
                {
                    'orderStrategyType': 'OCO',
                    'childOrderStrategies': [
                        {
                            'orderStrategyType': 'SINGLE',
                            'session': 'NORMAL',
                            'duration': 'GOOD_TILL_CANCEL',
                            'orderType': 'LIMIT',
                            'price': '15.27',
                            'orderLegCollection': [
                                {
                                    'instruction': 'SELL',
                                    'quantity': 5,
                                    'instrument': {
                                        'assetType': 'EQUITY',
                                        'symbol': 'XYZ'
                                    }
                                }
                            ]
                        },
                        {
                            'orderStrategyType': 'SINGLE',
                            'session': 'NORMAL',
                            'duration': 'GOOD_TILL_CANCEL',
                            'orderType': 'STOP',
                            'stopPrice': '11.27',
                            'orderLegCollection': [
                                {
                                    'instruction': 'SELL',
                                    'quantity': 5,
                                    'instrument': {
                                        'assetType': 'EQUITY',
                                        'symbol': 'XYZ'
                                    }
                                }
                            ]
                        }
                    ]
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))

    @no_duplicates
    def test_sell_trailing_stop_stock(self):
        builder = (
            OrderBuilder()
            .set_complex_order_strategy_type(ComplexOrderStrategyType.NONE)
            .set_order_type(OrderType.TRAILING_STOP)
            .set_session(Session.NORMAL)
            .set_stop_price_link_basis(StopPriceLinkBasis.BID)
            .set_stop_price_link_type(StopPriceLinkType.VALUE)
            .set_stop_price_offset(10)
            .set_duration(Duration.DAY)
            .set_order_strategy_type(OrderStrategyType.SINGLE)
            .add_equity_leg(EquityInstruction.SELL, 'XYZ', 10))

        expected = {
            'complexOrderStrategyType': 'NONE',
            'orderType': 'TRAILING_STOP',
            'session': 'NORMAL',
            'stopPriceLinkBasis': 'BID',
            'stopPriceLinkType': 'VALUE',
            'stopPriceOffset': 10,
            'duration': 'DAY',
            'orderStrategyType': 'SINGLE',
            'orderLegCollection': [
                {
                    'instruction': 'SELL',
                    'quantity': 10,
                    'instrument': {
                        'symbol': 'XYZ',
                        'assetType': 'EQUITY'
                    }
                }
            ]
        }

        self.assertFalse(has_diff(expected, builder.build()))


class TruncateFloatTest(unittest.TestCase):

    def test_zero(self):
        self.assertEqual('0.00', truncate_float(0))

    def test_zero_float(self):
        self.assertEqual('0.00', truncate_float(0.0))

    # positive numbers

    def test_integer(self):
        self.assertEqual('12.00', truncate_float(12))

    def test_integer_as_float(self):
        self.assertEqual('12.00', truncate_float(12.0))

    def test_three_digits(self):
        self.assertEqual('12.12', truncate_float(12.123))

    def test_three_digits_truncate_not_round(self):
        self.assertEqual('12.12', truncate_float(12.129))

    def test_less_than_one(self):
        self.assertEqual('0.1212', truncate_float(.12121))

    # same as above, except with negative numbers

    def test_negative_integer(self):
        self.assertEqual('-12.00', truncate_float(-12))

    def test_negative_integer_as_float(self):
        self.assertEqual('-12.00', truncate_float(-12.0))

    def test_negative_three_digits(self):
        self.assertEqual('-12.12', truncate_float(-12.123))

    def test_negative_three_digits_truncate_not_round(self):
        self.assertEqual('-12.12', truncate_float(-12.129))

    def test_negative_less_than_one(self):
        self.assertEqual('-0.1212', truncate_float(-.12121))



from colorama import Fore, Back, Style, init
from unittest.mock import AsyncMock, MagicMock

import asyncio
import difflib
import inspect
import httpx
import json

class AnyStringWith(str):
    '''
    Utility for checking whether a function was called with the given string as 
    a substring.
    '''
    def __eq__(self, other):
        return self in other

def account_preferences():
    with open('tests/testdata/preferences.json', 'r') as f:
        return json.load(f)


def real_order():
    return {
        'session': 'NORMAL',
        'duration': 'DAY',
        'orderType': 'MARKET',
        'complexOrderStrategyType': 'NONE',
        'quantity': 1.0,
        'filledQuantity': 1.0,
        'remainingQuantity': 0.0,
        'destinationLinkName': 'AUTO',
        'price': 58.41,
        'orderLegCollection': [
            {
                'orderLegType': 'EQUITY',
                'legId': 1,
                'instrument': {
                    'assetType': 'EQUITY',
                    'cusip': '126650100',
                    'symbol': 'CVS'
                },
                'instruction': 'BUY',
                'positionEffect': 'OPENING',
                'quantity': 1.0
            }
        ],
        'orderStrategyType': 'SINGLE',
        'orderId': 100001,
        'cancelable': False,
        'editable': False,
        'status': 'FILLED',
        'enteredTime': '2020-03-30T15:36:12+0000',
        'closeTime': '2020-03-30T15:36:12+0000',
        'tag': 'API_TDAM:App',
        'accountId': 100000,
        'orderActivityCollection': [
            {
                'activityType': 'EXECUTION',
                'executionType': 'FILL',
                'quantity': 1.0,
                'orderRemainingQuantity': 0.0,
                'executionLegs': [
                    {
                        'legId': 1,
                        'quantity': 1.0,
                        'mismarkedQuantity': 0.0,
                        'price': 58.1853,
                        'time': '2020-03-30T15:36:12+0000'
                    }
                ]
            }
        ]
    }

class AsyncResync:
    """
    Re-synchronizes every async function on a given object.
    NOTE: Every method runs on a new loop
    """

    class _AsyncResyncMethod:
        def __init__(self, func):
            self.func = func
        def __call__(self, *args, **kwargs):
            coroutine = self.func(*args, **kwargs)
            loop = asyncio.new_event_loop()
            retval = loop.run_until_complete(coroutine)
            loop.close()
            return retval

    def __getattr__(self, attr):
        retval = super().__getattribute__(attr)
        if inspect.iscoroutinefunction(retval):
            return self._AsyncResyncMethod(retval)
        return retval
    __getattribute__ = __getattr__


class ResyncProxy:
    """
    Proxies the underlying class, replacing coroutine methods
    with an auto-executing one
    """
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, *args, **kwargs):
        class DynamicResync(AsyncResync, self.cls):
            """
            Forces a mixin of the underlying class and the AsyncResync
            class
            """
            pass
        return DynamicResync(*args, **kwargs)

    def __getattr__(self, key):
        if key == 'cls':
            return super().__getattribute__(key)
        return getattr(self.cls, key)


# TODO: Figure out if httpx supports cleaner response mocking, because this is 
#       pretty janky.
class MockResponse(httpx.Response):
    def __init__(self, json, status_code, headers=None):
        resp_args = {
            'status_code': status_code,
            'json': json,
        }
        if headers:
            resp_args['headers'] = headers
        httpx.Response.__init__(
                self, **resp_args)


# TODO: This was necessary back before AsyncMock was a thing, but is it still 
#       necessary now?
class AsyncMagicMock:
    """
    Simple mock that returns an AsyncMock instance for every attribute. Useful
    for mocking async libraries.
    """
    def __init__(self):
        self.__attr_cache = {}

    def __getattr__(self, key):
        attr_cache = super().__getattribute__('_AsyncMagicMock__attr_cache')
        if key == '_AsyncMagicMock__attr_cache': return attr_cache

        try:
            return super().__getattribute__(key)
        except AttributeError:
            if key not in attr_cache:
                attr_cache[key] = AsyncMock()
            return attr_cache[key]

    def __setattr__(self, key, val):
        if key == '_AsyncMagicMock__attr_cache':
            return super().__setattr__(key, val)
        attr_cache = super().__getattribute__('_AsyncMagicMock__attr_cache')
        attr_cache[key] = val

    def __hasattr__(self, key):
        attr_cache = super().__getattribute__('_AsyncMagicMock__attr_cache')
        return attr_cache.has_key(key)

    def reset_mock(self):
        self.__attr_cache.clear()

def has_diff(old, new):
    old_out = json.dumps(old, indent=4, sort_keys=True).splitlines()
    new_out = json.dumps(new, indent=4, sort_keys=True).splitlines()
    diff = difflib.ndiff(old_out, new_out)
    diff, has_diff = color_diff(diff)

    if has_diff:
        print('\n'.join(diff))
    return has_diff


def color_diff(diff):
    has_diff = False
    output = []
    for line in diff:
        if line.startswith('+'):
            output.append(Fore.GREEN + line + Fore.RESET)
            has_diff = True
        elif line.startswith('-'):
            output.append(Fore.RED + line + Fore.RESET)
            has_diff = True
        elif line.startswith('^'):
            output.append(Fore.BLUE + line + Fore.RESET)
            has_diff = True
        else:
            output.append(line)
    return output, has_diff


__NO_DUPLICATES_DEFINED_NAMES = set()


def no_duplicates(f):
    name = f.__qualname__
    if name in __NO_DUPLICATES_DEFINED_NAMES:
        raise AttributeError('duplicate definition of {}'.format(name))
    __NO_DUPLICATES_DEFINED_NAMES.add(name)
    return f


class MockOAuthClient(MagicMock):
    def __call__(self, *args, **kwargs):
        if 'update_token' in kwargs:
            assert not inspect.iscoroutinefunction(kwargs['update_token'])
        return MagicMock.__call__(self, *args, **kwargs)


class MockAsyncOAuthClient(MagicMock):
    def __call__(self, *args, **kwargs):
        if 'update_token' in kwargs:
            assert inspect.iscoroutinefunction(kwargs['update_token'])
        return MagicMock.__call__(self, *args, **kwargs)



import argparse
import httpx
import json

from schwab.auth import client_from_token_file
from schwab.contrib.orders import construct_repeat_order, code_for_builder


def latest_order_main(sys_args):
    parser = argparse.ArgumentParser(
            description='Utilities for generating code from historical orders')

    required = parser.add_argument_group('required arguments')
    required.add_argument(
            '--token_file', required=True, help='Path to token file')
    required.add_argument('--api_key', required=True)
    required.add_argument('--app_secret', required=True)

    account_spec_group = parser.add_mutually_exclusive_group()
    account_spec_group.add_argument('--account_id', type=int,
            help='Restrict lookups to a specific account ID')

    account_spec_group.add_argument('--account_hash', type=str,
            help='Restrict lookups to the account with the specified hash')

    args = parser.parse_args(args=sys_args)
    client = client_from_token_file(
            args.token_file, args.app_secret, args.api_key)

    # If the account ID is specified, find the corresponding account hash
    if args.account_id is not None:
        r = client.get_account_numbers()
        assert r.status_code == httpx.codes.OK

        for val in r.json():
            if val['accountNumber'] == str(args.account_id):
                account_hash = val['hashValue']
                break
        else:
            print(('Failed to find account has for account ID {}. Searched ' +
                   'the following accounts:\n{}').format(
                       args.account_id, json.dumps(r.json(), indent=4)))
            return -1
    else:
        account_hash = args.account_hash


    # Fetch orders
    def get_orders(method):
        r = method()
        if r.status_code != httpx.codes.OK:
            print(('Returned HTTP status code {}. This is most often caused ' +
                   'by an invalid account ID or hash.').format(r.status_code))
            return None
        return r.json()

    if account_hash is not None:
        orders = get_orders(lambda: client.get_orders_for_account(account_hash))
        if orders is None:
            return -1

        if 'error' in orders:
            print(('Schwab returned error: "{}", This is most often caused ' +
                   'by an invalid account ID or hash').format(orders['error']))
            return -1
    else:
        orders = get_orders(lambda: client.get_orders_for_all_linked_accounts())
        if orders is None:
            return -1

        if 'error' in orders:
            print('Schwab returned error: "{}"'.format(orders['error']))
            return -1

    # Construct and emit order code
    if orders:
        order = sorted(orders, key=lambda o: -o['orderId'])[0]

        # XXX: If necessary, warn about the jank around destinations
        emit_destination_warning_newline = False
        if ('requestedDestination' in order
                and order['requestedDestination'] != 'AUTO'):
            print(('# Warning: This order contains a non-"AUTO" value of ' +
                   '"requestedDestination" ("{}").').format(
                       order['requestedDestination']))
            print('#          This parameter appears to be broken in ' +
                  'the API, so it is omitted in this generated code.')
            emit_destination_warning_newline = True
        if ('destinationLinkName' in order
                and order['destinationLinkName'] != 'AutoRoute'):
            print(('# Warning: This order contains a non-"AutoRoute" value of ' +
                   '"destinationLinkName" ("{}").').format(
                           order['destinationLinkName']))
            print('           This parameter appears to be broken in the ' +
                  'API, so it is omitted in this generated code.''')
            emit_destination_warning_newline = True
        if emit_destination_warning_newline:
            print()

        print('# Order ID', order['orderId'])
        print(code_for_builder(construct_repeat_order(order)))
    else:
        print('No recent orders found')

    return 0



import callee
import httpx
import subprocess
import unittest
from unittest.mock import call, MagicMock, patch

from ..utils import AnyStringWith, no_duplicates
from schwab.scripts.orders_codegen import latest_order_main

class LatestOrderTest(unittest.TestCase):

    def setUp(self):
        self.args = []

    def add_arg(self, arg):
        self.args.append(arg)

    def main(self):
        return latest_order_main(self.args)

    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_success_no_account_id(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')

        orders = [
                {'orderId': 201},
                {'orderId': 101},
                {'orderId': 301},
                {'orderId': 401},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_all_linked_accounts.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_called_once_with(orders[3])
        mock_print.assert_has_calls([
                call('# Order ID', 401),
                call(mock_code_for_builder.return_value)])


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_no_account_id_no_recent_orders(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')

        orders = []

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_all_linked_accounts.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_not_called()
        mock_print.assert_called_once_with('No recent orders found')


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_no_account_id_error(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')

        orders = {'error': 'invalid'}

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_all_linked_accounts.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), -1)

        mock_construct_repeat_order.assert_not_called()
        mock_print.assert_called_once_with(
                AnyStringWith('Schwab returned error: "invalid"'))


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_account_id_success(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_id')
        self.add_arg('123456')

        orders = [
                {'orderId': 201},
                {'orderId': 101},
                {'orderId': 301},
                {'orderId': 401},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_account_numbers.return_value = httpx.Response(
                200,
                json=[{
                    'accountNumber': '123456',
                    'hashValue': 'hash-value',
                }])
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_client.get_account_numbers.assert_called_once()
        mock_construct_repeat_order.assert_called_once_with(orders[3])
        mock_print.assert_has_calls([
                call('# Order ID', 401),
                call(mock_code_for_builder.return_value)])


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_account_id_no_corresponding_hash(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_id')
        self.add_arg('123456')

        orders = [
                {'orderId': 201},
                {'orderId': 101},
                {'orderId': 301},
                {'orderId': 401},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_account_numbers.return_value = httpx.Response(
                200,
                json=[{
                    'accountNumber': '90009',
                    'hashValue': 'hash-value',
                }])
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), -1)

        mock_client.get_account_numbers.assert_called_once()
        mock_print.assert_called_once_with(
                AnyStringWith('Failed to find account has for account ID'))


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_account_id_no_orders(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_id')
        self.add_arg('123456')

        orders = []

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_account_numbers.return_value = httpx.Response(
                200,
                json=[{
                    'accountNumber': '123456',
                    'hashValue': 'hash-value',
                }])
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_not_called
        mock_print.assert_called_once_with('No recent orders found')


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_account_id_error(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_id')
        self.add_arg('123456')

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_account_numbers.return_value = httpx.Response(
                200,
                json=[{
                    'accountNumber': '123456',
                    'hashValue': 'hash-value',
                }])
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json={'error': 'invalid'})

        self.assertEqual(self.main(), -1)

        mock_construct_repeat_order.assert_not_called
        mock_print.assert_called_once_with(
                AnyStringWith('Schwab returned error: "invalid"'))


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_success_account_hash(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_hash')
        self.add_arg('account-hash')

        orders = [
                {'orderId': 201},
                {'orderId': 101},
                {'orderId': 301},
                {'orderId': 401},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_called_once_with(orders[3])
        mock_print.assert_has_calls([
                call('# Order ID', 401),
                call(mock_code_for_builder.return_value)])


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_order_fetching_fails_no_account_id(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_all_linked_accounts.return_value \
                = httpx.Response(400)

        self.assertEqual(self.main(), -1)

        mock_print.assert_called_once_with(
                AnyStringWith('Returned HTTP status code 400'))


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_order_fetching_fails_account_id(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_id')
        self.add_arg('123456')

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_account_numbers.return_value = httpx.Response(
                200,
                json=[{
                    'accountNumber': '123456',
                    'hashValue': 'hash-value',
                }])
        mock_client.get_orders_for_account.return_value = httpx.Response(400)

        self.assertEqual(self.main(), -1)

        mock_print.assert_called_once_with(
                AnyStringWith('Returned HTTP status code 400'))


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_warn_on_non_auto_requestedDestination(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_hash')
        self.add_arg('account-hash')

        orders = [
                {'orderId': 401,
                 'requestedDestination': 'not AUTO'},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_called_once_with(orders[0])
        mock_print.assert_has_calls([
                call(callee.Contains('requestedDestination')),
                call(callee.Contains('broken')),
                call(),
                call('# Order ID', 401),
                call(mock_code_for_builder.return_value)])


    @no_duplicates
    @patch('builtins.print')
    @patch('schwab.scripts.orders_codegen.client_from_token_file')
    @patch('schwab.scripts.orders_codegen.construct_repeat_order')
    @patch('schwab.scripts.orders_codegen.code_for_builder')
    def test_warn_on_non_auto_destinationLinkName(
            self,
            mock_code_for_builder,
            mock_construct_repeat_order,
            mock_client_from_token_file,
            mock_print):
        self.add_arg('--token_file')
        self.add_arg('filename.json')
        self.add_arg('--api_key')
        self.add_arg('api-key')
        self.add_arg('--app_secret')
        self.add_arg('app-secret')
        self.add_arg('--account_hash')
        self.add_arg('account-hash')

        orders = [
                {'orderId': 401,
                 'destinationLinkName': 'not AUTO'},
        ]

        mock_client = MagicMock()
        mock_client_from_token_file.return_value = mock_client
        mock_client.get_orders_for_account.return_value \
                = httpx.Response(200, json=orders)

        self.assertEqual(self.main(), 0)

        mock_construct_repeat_order.assert_called_once_with(orders[0])
        mock_print.assert_has_calls([
                call(callee.Contains('destinationLinkName')),
                call(callee.Contains('broken')),
                call(),
                call('# Order ID', 401),
                call(mock_code_for_builder.return_value)])


class ScriptInvocationTest(unittest.TestCase):

    def test_get_help(self):
        output = subprocess.check_output(
                'schwab-order-codegen.py --help',
                shell=True, text=True)



import autopep8
import schwab

from schwab.orders.generic import OrderBuilder
from schwab.orders.common import (
        EquityInstrument,
        OptionInstrument,
        OrderStrategyType,
)

from collections import defaultdict


def _call_setters_with_values(order, builder):
    '''
    For each field in fields_and_setters, if it exists as a key in the order
    object, pass its value to the appropriate setter on the order builder.
    '''
    for field_name, setter_name, enum_class in _FIELDS_AND_SETTERS:
        try:
            value = order[field_name]
        except KeyError:
            continue

        if enum_class:
            value = enum_class[value]

        setter = getattr(builder, setter_name)
        setter(value)


# Top-level fields
_FIELDS_AND_SETTERS = (
    ('session', 'set_session', schwab.orders.common.Session),
    ('duration', 'set_duration', schwab.orders.common.Duration),
    ('orderType', 'set_order_type', schwab.orders.common.OrderType),
    ('complexOrderStrategyType', 'set_complex_order_strategy_type',
        schwab.orders.common.ComplexOrderStrategyType),
    ('quantity', 'set_quantity', None),
    # XXX: Destinations are weird/busted
    #       * As of 2024-05-16, the example in the place_order documentation 
    #         references destinationLinkName but not requestedDestination.
    #       * The same documentation lists requestedDestination as a parameter 
    #         to the method, but doesn't list destinationLinkName.
    #       * Fetching historical orders returns orders which contain both.
    #      These parameters are being ignored until we gain more clarity.
    #('destinationLinkName', 'set_destination_link_name',
    #    schwab.orders.common.Destination),
    ('stopPrice', 'copy_stop_price', None),
    ('stopPriceLinkBasis', 'set_stop_price_link_basis',
        schwab.orders.common.StopPriceLinkBasis),
    ('stopPriceLinkType', 'set_stop_price_link_type',
        schwab.orders.common.StopPriceLinkType),
    ('stopPriceOffset', 'set_stop_price_offset', None),
    ('stopType', 'set_stop_type', schwab.orders.common.StopType),
    ('priceLinkBasis', 'set_price_link_basis',
        schwab.orders.common.PriceLinkBasis),
    ('priceLinkType', 'set_price_link_type',
        schwab.orders.common.PriceLinkType),
    ('price', 'copy_price', None),
    ('activationPrice', 'set_activation_price', None),
    ('specialInstruction', 'set_special_instruction',
        schwab.orders.common.SpecialInstruction),
    ('orderStrategyType', 'set_order_strategy_type',
        schwab.orders.common.OrderStrategyType),
)

def construct_repeat_order(historical_order):
    builder = schwab.orders.generic.OrderBuilder()

    # Top-level fields
    _call_setters_with_values(historical_order, builder)

    # Composite orders
    if 'orderStrategyType' in historical_order:
        if historical_order['orderStrategyType'] == 'TRIGGER':
            builder = schwab.orders.common.first_triggers_second(
                    builder, construct_repeat_order(
                        historical_order['childOrderStrategies'][0]))
        elif historical_order['orderStrategyType'] == 'OCO':
            builder = schwab.orders.common.one_cancels_other(
                    construct_repeat_order(
                        historical_order['childOrderStrategies'][0]),
                    construct_repeat_order(
                        historical_order['childOrderStrategies'][1]))
    else:
        raise ValueError('historical order is missing orderStrategyType')

    # Order legs
    if 'orderLegCollection' in historical_order:
        for leg in historical_order['orderLegCollection']:
            if leg['orderLegType'] == 'EQUITY':
                builder.add_equity_leg(
                        schwab.orders.common.EquityInstruction[leg['instruction']],
                        leg['instrument']['symbol'],
                        leg['quantity'])
            elif leg['orderLegType'] == 'OPTION':
                builder.add_option_leg(
                        schwab.orders.common.OptionInstruction[leg['instruction']],
                        leg['instrument']['symbol'],
                        leg['quantity'])
            else:
                raise ValueError(
                        'unknown orderLegType {}'.format(leg['orderLegType']))

    return builder


################################################################################
# AST generation


def code_for_builder(builder, var_name=None):
    '''
    Returns code that can be executed to construct the given builder, including
    import statements.

    :param builder: :class:`~schwab.orders.generic.OrderBuilder` to generate.
    :param var_name: If set, emit code that assigns the builder to a variable
                     with this name.
    '''
    ast = construct_order_ast(builder)

    imports = defaultdict(set)
    lines = []
    ast.render(imports, lines)

    import_lines = []
    for module, names in imports.items():
        line = 'from {} import {}'.format(
                module, ', '.join(names))
        if len(line) > 80:
            line = 'from {} import (\n{}\n)'.format(
                    module, ',\n'.join(names))
        import_lines.append(line)

    if var_name:
        var_prefix = f'{var_name} = '
    else:
        var_prefix = ''

    return autopep8.fix_code(
            '\n'.join(import_lines) + 
            '\n\n' +
            var_prefix +
            '\n'.join(lines))


class FirstTriggersSecondAST:
    def __init__(self, first, second):
        self.first = first
        self.second = second

    def render(self, imports, lines, paren_depth=0):
        imports['schwab.orders.common'].add('first_triggers_second')

        lines.append('first_triggers_second(')
        self.first.render(imports, lines, paren_depth + 1)
        lines[-1] += ','
        self.second.render(imports, lines, paren_depth + 2)
        lines.append(')')


class OneCancelsOtherAST:
    def __init__(self, one, other):
        self.one = one
        self.other = other

    def render(self, imports, lines, paren_depth=0):
        imports['schwab.orders.common'].add('one_cancels_other')

        lines.append('one_cancels_other(')
        self.one.render(imports, lines, paren_depth + 1)
        lines[-1] += ','
        self.other.render(imports, lines, paren_depth + 2)
        lines.append(')')


class FieldAST:
    def __init__(self, setter_name, enum_type, value):
        self.setter_name = setter_name
        self.enum_type = enum_type
        self.value = value

    def render(self, imports, lines, paren_depth=0):
        value = self.value
        if self.enum_type:
            imports[self.enum_type.__module__].add(self.enum_type.__qualname__)
            value = self.enum_type.__qualname__ + '.' + value

        lines.append(f'.{self.setter_name}({value})')


class EquityOrderLegAST:
    def __init__(self, instruction, symbol, quantity):
        self.instruction = instruction
        self.symbol = symbol
        self.quantity = quantity

    def render(self, imports, lines, paren_depth=0):
        imports['schwab.orders.common'].add('EquityInstruction')
        lines.append('.add_equity_leg(EquityInstruction.{}, "{}", {})'.format(
            self.instruction, self.symbol, self.quantity))


class OptionOrderLegAST:
    def __init__(self, instruction, symbol, quantity):
        self.instruction = instruction
        self.symbol = symbol
        self.quantity = quantity

    def render(self, imports, lines, paren_depth=0):
        imports['schwab.orders.common'].add('OptionInstruction')
        lines.append('.add_option_leg(OptionInstruction.{}, "{}", {})'.format(
            self.instruction, self.symbol, self.quantity))


class GenericBuilderAST:
    def __init__(self, builder):
        self.top_level_fields = []
        for name, setter, enum_type in sorted(_FIELDS_AND_SETTERS):
            value = getattr(builder, '_'+name)
            if value is not None:
                self.top_level_fields.append(FieldAST(setter, enum_type, value))

        for leg in builder._orderLegCollection:
            if leg['instrument']._assetType == 'EQUITY':
                self.top_level_fields.append(EquityOrderLegAST(
                    leg['instruction'], leg['instrument']._symbol, 
                    leg['quantity']))
            elif leg['instrument']._assetType == 'OPTION':
                self.top_level_fields.append(OptionOrderLegAST(
                    leg['instruction'], leg['instrument']._symbol, 
                    leg['quantity']))
            else:
                raise ValueError('unknown leg asset type {}'.format(
                    leg['instrument']._assetType))

    def render(self, imports, lines, paren_depth=0):
        imports['schwab.orders.generic'].add('OrderBuilder')

        lines.append('OrderBuilder() \\')
        for idx, field in enumerate(self.top_level_fields):
            field.render(imports, lines, paren_depth)

            if paren_depth == 0 and idx != len(self.top_level_fields) - 1:
                lines[-1] += ' \\'


def construct_order_ast(builder):
    if builder._orderStrategyType == 'OCO':
        return OneCancelsOtherAST(
                construct_order_ast(builder._childOrderStrategies[0]),
                construct_order_ast(builder._childOrderStrategies[1]))
    elif builder._orderStrategyType == 'TRIGGER':
        return FirstTriggersSecondAST(
                GenericBuilderAST(builder),
                construct_order_ast(builder._childOrderStrategies[0]))
    else:
        return GenericBuilderAST(builder)
